use std::sync::{Arc, OnceLock, RwLock};

use jni::{
  objects::{JByteBuffer, JIntArray, JObject, JString, JValue},
  signature::{Primitive, ReturnType},
  strings::JNIString,
  sys::{jint, jlong},
  JNIEnv,
};

use crate::{
  assert_positive,
  component::{mmio::ComponentMmioControl, Component, ComponentInfo, ComponentInterface},
  core::CoreHandle,
  unsync_cell::UnsyncCell,
};

pub struct Computer {
  pub components: Arc<[RwLock<Component>]>,
  pub component_control: Arc<UnsyncCell<ComponentMmioControl>>,
  pub shared_ram: Arc<UnsyncCell<[u8]>>,
  pub shared_rom: Arc<UnsyncCell<[u8]>>,

  pub cores: OnceLock<Box<[CoreHandle]>>,
}

impl Computer {
  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_new")]
  extern "C" fn new<'local>(
    mut env: JNIEnv<'local>,
    java_computer: JObject<'local>,
    component_capacity: i32,
    memory_size: i32,
    core_speeds: JIntArray<'local>,
  ) {
    let (component_capacity, memory_size): (usize, usize) = (
      assert_positive!(env, component_capacity),
      assert_positive!(env, memory_size),
    );

    let core_count = env.get_array_length(&core_speeds).unwrap() as usize;
    if core_count > 16 {
      env.throw("no more than 16 cores are allowed").unwrap();
      return;
    }
    println!("got to computer construction!");
    let computer = Box::new(Self {
      components: {
        let mut components = Vec::with_capacity(component_capacity);
        components.extend(std::iter::from_fn(|| Default::default()).take(component_capacity));
        Arc::from(components.into_boxed_slice())
      },
      component_control: Arc::new(UnsyncCell::new(ComponentMmioControl::new())),
      shared_ram: Arc::from(UnsyncCell::new_zeroed_box(memory_size as usize)),
      shared_rom: Arc::from(UnsyncCell::new_zeroed_box(0x1000000 as usize)),
      // late initalized, see below
      cores: OnceLock::new(),
    });

    println!("core construction");
    let mut speeds = [0; 16];
    env.get_int_array_region(core_speeds, 0, &mut speeds[..core_count]).unwrap();
    println!("core speeds acquired {:?}", &speeds[..core_count]);
    let Ok(cores) = speeds
      .into_iter()
      .take(core_count)
      .enumerate()
      .map(|(core_index, speed)| {
        Ok(CoreHandle::new(
          &computer,
          core_index,
          assert_positive!(env, speed, Err(())),
          env.get_java_vm().unwrap(),
        ))
      })
      .collect::<Result<Vec<_>, _>>()
    else {
      return;
    };

    let _ = computer.cores.set(cores.into_boxed_slice());

    println!("storing computer info in field!");

    env
      .set_field(
        &java_computer,
        "address",
        "J",
        JValue::Long(Box::into_raw(computer) as jlong),
      )
      .unwrap();
  }

  fn from_jni<'local>(env: &mut JNIEnv<'local>, computer: JObject<'local>) -> *mut Self {
    env
      .get_field_unchecked(
        &computer,
        (env.get_object_class(&computer).unwrap(), "address", "J"),
        ReturnType::Primitive(Primitive::Long),
      )
      .unwrap()
      .j()
      .unwrap() as *mut _
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_insertComponentNative")]
  unsafe extern "C" fn insert_component<'local>(
    mut env: JNIEnv<'local>,
    computer: JObject<'local>,
    component_index: jint,
    component_object: JObject<'local>,
    type_name: JString<'local>,
    protocol: JString<'local>,
    bytes_per_tick: u32,
    version: u32,
  ) {
    // Safety: the responsibility of asserting the pointer is valid is on the java caller
    let computer = unsafe { &*Self::from_jni(&mut env, computer) };

    let component_index: usize = assert_positive!(env, component_index);
    let mut make_info_string = |string: JString<'local>| {
      let mut bytes = [0u8; 256];
      let jstring = env.get_string(&string).unwrap();

      let length = bytes.len().min(jstring.to_bytes().len());
      bytes[..length].copy_from_slice(&jstring.to_bytes()[..length]);

      bytes
    };
    let type_name = make_info_string(type_name);
    let protocol = make_info_string(protocol);
    let info = ComponentInfo {
      type_name,
      protocol,
      bytes_per_tick,
      version,
    };
    let java_component = env.new_global_ref(component_object).unwrap();
    *computer.components[component_index].write().unwrap() = Component::Populated {
      own_index: component_index,
      java_component,
      read_transfer_active: false.into(),
      write_transfer_active: false.into(),
      info,
      interface: ComponentInterface::default().into(),
    };
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_updateRomTest")]
  unsafe extern "C" fn update_rom_test<'local>(
    mut env: JNIEnv<'local>,
    computer: JObject<'local>,
    bytes: JByteBuffer<'local>,
  ) {
    // Safety: the responsibility of asserting the pointer is valid is on the java caller
    let computer = unsafe { &*Self::from_jni(&mut env, computer) };

    computer.shared_rom.update_from_jni(&mut env, &bytes).expect("failed to update rom");
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_updateRomFile")]
  unsafe extern "C" fn update_rom_file<'local>(env: JNIEnv<'local>, computer: JObject<'local>) {
    drop((env, computer));
    todo!()
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_free")]
  unsafe extern "C" fn free<'local>(mut env: JNIEnv<'local>, computer: JObject<'local>) {
    let computer = Self::from_jni(&mut env, computer);
    // Safety: the responsibility of asserting the pointer is valid is on the java caller
    drop(unsafe { Box::from_raw(computer) })
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_wakeCoreAt")]
  unsafe extern "C" fn wake_core_at<'local>(
    mut env: JNIEnv<'local>,
    computer: JObject<'local>,
    core: jint,
    address: jlong,
  ) {
    let (core, address): (usize, u64) = (assert_positive!(env, core), assert_positive!(env, address));
    // Safety: the responsibility of asserting the pointer is valid is on the java caller
    let computer = unsafe { &*Self::from_jni(&mut env, computer) };

    let Some(core) = computer.cores.get().unwrap().get(core) else {
      env.throw("core index was out of bounds").unwrap();
      return;
    };
    core.wake_at(address);
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_tick")]
  unsafe extern "C" fn tick<'local>(mut env: JNIEnv<'local>, computer: JObject<'local>) {
    // Safety: the responsibility of asserting the pointer is valid is on the java caller
    let computer = unsafe { &*Self::from_jni(&mut env, computer) };
    let cores = &computer.cores.get().unwrap();
    for core in cores.iter() {
      core.tick();
    }
  }
}
